#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>

#include "shfs.h"
#include "shfs_dcache.h"
#include "shfs_proc.h"

#define dir_cache info->dir_cache

void
shfs_dcache_init(struct shfs_sb_info *info) {
	memset(&dir_cache, 0, sizeof(struct shfs_dir_cache));
}

unsigned long
shfs_dcache_hash(char *name) {
	unsigned long hash = 0;
	int i;

	for (i = 0; i < strlen(name); i++)
		hash += name[i];
	return hash % SHFS_DCACHE_HASH;
}

int
shfs_dcache_deldir(struct shfs_sb_info *info, struct shfs_dir_hash_node *n) {
	struct shfs_dir_list_node *p, *q;
	unsigned long hsh;

	if (!n) 
		return -1;
	
	DEBUG("deleting a cache entry (%s)\n", n->directory.name);
	hsh = shfs_dcache_hash(n->directory.name);
	
	for (p = n->directory.head; p; p = q) {
		q = p->next;
		KMEM_FREE("name", dir_name_cache, p->entry.name);
		KMEM_FREE("dir", dir_list_cache, p);
	}

	KMEM_FREE("name", dir_name_cache, n->directory.name);
	if (n->prev)
		n->prev->next = n->next;
	if (n->next)
		n->next->prev = n->prev;
	if (n == dir_cache.hash[hsh])
		dir_cache.hash[hsh] = n->next;

	KMEM_FREE("hash", dir_hash_cache, n);
	dir_cache.len[hsh]--;

	DEBUG("done\n");
	return 0;
}

int
shfs_dcache_empty(struct shfs_sb_info *info)
{
	int i;

	DEBUG("closing cache...\n");
	for (i = 0; i < SHFS_DCACHE_HASH; i++) {
		DEBUG("closing hash bucket %d (%d)...\n", i, dir_cache.len[i]);
		while (dir_cache.len[i])
			shfs_dcache_deldir(info, dir_cache.hash[i]);
	}
	DEBUG(" OK\n");
	return 0;
}

int
shfs_dcache_infront(struct shfs_sb_info *info, struct shfs_dir_hash_node *n)
{
	unsigned long hsh;

	DEBUG("bringing in front %s\n", n->directory.name);
	if (n->prev) {
		hsh = shfs_dcache_hash(n->directory.name);
			
		n->prev->next = n->next;

		if (n->next)
			n->next->prev = n->prev;

		n->prev = NULL;
		n->next = dir_cache.hash[hsh];
		if (n->next)
			n->next->prev = n;
		dir_cache.hash[hsh] = n;
	}
	return 0;
}


int
shfs_dcache_shrink(unsigned long hsh, struct shfs_sb_info *info)
{
	struct shfs_dir_hash_node *p;

	DEBUG("Shrinking hash bucket %u (%d)\n", (unsigned)hsh, dir_cache.len[hsh]);

	if (dir_cache.len[hsh] == 0)
		return -1;
	
	for (p = dir_cache.hash[hsh]; p->next != NULL; p = p->next)
		;
	shfs_dcache_deldir(info, p);
	
	return 0;
}

int
shfs_dcache_add(struct shfs_sb_info *info, char *name, struct shfs_directory **dir)
{
	unsigned long hsh;
	struct shfs_dir_hash_node *p;
	int res;
	
	hsh = shfs_dcache_hash(name);
	
	while (dir_cache.len[hsh] >= SHFS_DCACHE_LEN)
		shfs_dcache_shrink(hsh, info);
	p = (struct shfs_dir_hash_node*)KMEM_ALLOC("hash", dir_hash_cache, GFP_KERNEL);
	if (!p) {
		VERBOSE("slab cache error!\n");
		res = -ENOMEM;
		goto out;
	}

	memset(p, 0, sizeof(struct shfs_dir_hash_node));
	p->directory.time = CURRENT_TIME;
	if (strlen(name) + 1 > SHFS_PATH_MAX) {
		VERBOSE("too long dir entry!\n");
		res = -ENAMETOOLONG;
		goto out1;
	}
	
	p->directory.name = (char *)KMEM_ALLOC("name", dir_name_cache, GFP_KERNEL);
	if (!p->directory.name) {
		VERBOSE("slab cache error!\n");
		res = -ENOMEM;
		goto out1;
	}

	strcpy(p->directory.name, name);

	DEBUG("loading dir...\n");
	if ((res = shfs_loaddir(info, name, &p->directory)) < 0) {
		VERBOSE("Couldn't load directory!\n");
		goto out2;
	}
	p->prev = NULL;
	p->next = dir_cache.hash[hsh];
	if (p->next)
		p->next->prev = p;
	dir_cache.hash[hsh] = p;

	dir_cache.len[hsh]++;
	*dir = &p->directory;
	return 0;
out2:
	KMEM_FREE("name", dir_name_cache, p->directory.name);
out1:
	KMEM_FREE("hash", dir_hash_cache, p);
out:
	return res;
}

int
shfs_dcache_get(struct shfs_sb_info *info, char *name, struct shfs_directory **dir)
{
	struct shfs_dir_hash_node *p;
	unsigned long hsh;

	DEBUG("looking for %s\n", name);
	hsh = shfs_dcache_hash(name);
	for (p = dir_cache.hash[hsh]; p; p = p->next) {
		if (!strcmp(name, p->directory.name)) {
			DEBUG(" found in cache!\n");
			if (CURRENT_TIME - p->directory.time > SHFS_DCACHE_TTL) {
				DEBUG(" cache entry too old!\n");
				if (shfs_dcache_deldir(info, p) < 0)
					VERBOSE("cannot delete dir!\n");
				return shfs_dcache_add(info, name, dir);
			}
			shfs_dcache_infront(info, p);
			*dir = &p->directory;
			return 0;
		}
	}
	DEBUG(" %s not found in cache. Adding...!\n", name);
	return shfs_dcache_add(info, name, dir);
}

int
shfs_dcache_revalidate(struct dentry *dentry, struct shfs_sb_info *info)
{
	char buf[SHFS_PATH_MAX];
	struct shfs_dir_hash_node *p;
	struct shfs_dir_list_node *q;
	struct shfs_directory *dir;
	unsigned long hsh;

	if (shfs_get_name(dentry->d_parent, buf) < 0)
		return 0;
	
	DEBUG("%s\n", buf);
	hsh = shfs_dcache_hash(buf);
	for (p = dir_cache.hash[hsh]; p != NULL; p = p->next) {
		if (!strcmp(buf, p->directory.name)) {
			if (CURRENT_TIME - p->directory.time > SHFS_DCACHE_TTL)
				goto revalidate;
			strcpy(buf, dentry->d_name.name);
			for (q = p->directory.head; q; q = q->next) {
				if (!strcmp(buf, q->entry.name)) 
					return 1;
			}
revalidate:
			if (shfs_dcache_deldir(info, p) < 0)
				VERBOSE("cannot delete dir!\n");
			return (shfs_dcache_add(info, buf, &dir) == 0);
		}
	}
	return (shfs_dcache_add(info, buf, &dir) == 0);
}

void
shfs_dcache_invalidate(struct dentry *dentry)
{
	struct super_block *sb = dentry->d_inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	char buf[SHFS_PATH_MAX];
	struct shfs_dir_hash_node *p;
	unsigned long hsh;
		
	if (shfs_get_name(dentry, buf) < 0)
		return;
	hsh = shfs_dcache_hash(buf);
	DEBUG(" \n");
	
	for (p = dir_cache.hash[hsh]; p != NULL; p = p->next) {
		if (!strcmp(buf, p->directory.name)) {
			shfs_dcache_deldir(info, p);
			return;
		}
	}
}

void
shfs_dcache_update_fattr(struct dentry *dentry, struct iattr *attr)
{
	struct super_block *sb = dentry->d_inode->i_sb;
	struct shfs_sb_info *info = (struct shfs_sb_info*)sb->u.generic_sbp;
	char buf[SHFS_PATH_MAX];
	struct shfs_dir_hash_node *p;
	struct shfs_dir_list_node *q;
	unsigned long hsh;
	
	if (shfs_get_name(dentry->d_parent, buf) < 0)
		return;
	hsh = shfs_dcache_hash(buf);
	for (p = dir_cache.hash[hsh]; p; p = p->next) {
		if (!strcmp(buf, p->directory.name)) {
			strcpy(buf, dentry->d_name.name);
			DEBUG("%s\n", buf);
			for (q = p->directory.head; q; q = q->next) {
				if (!strcmp(buf, q->entry.name)) {
					if (attr->ia_valid & ATTR_MODE)
						q->entry.mode = attr->ia_mode;
					if (attr->ia_valid & ATTR_UID)
						q->entry.uid = attr->ia_uid;
					if (attr->ia_valid & ATTR_GID)
						q->entry.gid = attr->ia_gid;
					if (attr->ia_valid & ATTR_SIZE) {
						q->entry.size = attr->ia_size;
						q->entry.blocksize = 4096;
						q->entry.blocks = (attr->ia_size + 4095) >> 11;
						q->entry.atime = CURRENT_TIME;
						DEBUG("size changed to: %llu\n", attr->ia_size);
					}
					return;
				}
			}
			shfs_dcache_deldir(info, p);
			return;
		}
	}
}

